import json
import os
import pathlib
import requests
import shutil
from flask import request
from http import HTTPStatus
from typing import Optional, Dict, Any
from urllib.parse import urljoin

from .base_code_block import BaseCodeBlock


class BlobStorage(BaseCodeBlock):
    """
    This class exposes Bismuth's blob storage service, allowing data to be persisted.
    Bytes are stored plain, while other types are stored as JSON.
    When run locally, it stores data via files in a temporary directory.
    """

    def __init__(self, api_url="http://169.254.169.254:9000/blob/v1/"):
        """
        Initialize the datastore.
        """
        if 'BISMUTH_AUTH' in os.environ:
            self._impl = _HostedBlobStorage(os.environ['BISMUTH_AUTH'], api_url)
        else:
            self._impl = _LocalBlobStorage()

    def _encode(self, value) -> bytes:
        return value if isinstance(value, bytes) else json.dumps(value).encode('utf-8')

    def _decode(self, value) -> Optional[Any]:
        if value is None:
            return None
        try:
            return json.loads(value)
        except (UnicodeDecodeError, json.JSONDecodeError):
            return value

    def create(self, key: str, value: Any) -> None:
        """Create a new item in the datastore."""
        return self._impl.create(key, self._encode(value))

    def retrieve(self, key: str) -> Optional[Any]:
        """Retrieve an item from the datastore."""
        return self._decode(self._impl.retrieve(key))

    def update(self, key: str, value: Any) -> None:
        """Update an existing item in the datastore."""
        return self._impl.update(key, self._encode(value))

    def delete(self, key: str) -> None:
        """Delete an item from the datastore."""
        return self._impl.delete(key)

    def list_keys(self) -> set[str]:
        """List all keys in the datastore."""
        return self._impl.list_keys()


class _HostedBlobStorage(BaseCodeBlock):
    # A dictionary of HTTP headers used to authenticate to the storage backend.
    _auth: Dict[str, str]
    # The URL of the storage backend.
    _api_url: str

    def __init__(self, auth: str, api_url="http://169.254.169.254:9000/blob/v1/"):
        self._auth = {"Authorization": "Bearer " + auth}
        self._api_url = api_url

    def _headers(self):
        hdrs = self._auth.copy()
        try:
            for tracehdr in ["traceparent", "tracestate"]:
                if tracehdr in request.headers:
                    hdrs[tracehdr] = request.headers[tracehdr]
        except RuntimeError:
            pass
        return hdrs

    def create(self, key: str, value: bytes) -> None:
        resp = requests.post(urljoin(self._api_url, key), data=value, headers=self._headers())
        if resp.status_code == HTTPStatus.CONFLICT:
            raise ValueError("Key already exists.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def retrieve(self, key: str) -> Optional[bytes]:
        resp = requests.get(urljoin(self._api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            return None
        elif not resp.ok:
            raise Exception(f"Server error {resp}")
        return resp.content

    def update(self, key: str, value: bytes) -> None:
        resp = requests.put(urljoin(self._api_url, key), data=value, headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError("Key does not exist.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def delete(self, key: str) -> None:
        resp = requests.delete(urljoin(self._api_url, key), headers=self._headers())
        if resp.status_code == HTTPStatus.NOT_FOUND:
            raise ValueError("Key does not exist.")
        elif not resp.ok:
            raise Exception(f"Server error {resp}")

    def list_keys(self) -> set[str]:
        resp = requests.get(self._api_url, headers=self._headers())
        if not resp.ok:
            raise Exception(f"Server error {resp}")
        return set(resp.json().keys())


class _LocalBlobStorage(BaseCodeBlock):
    _dir = pathlib.Path("/tmp/bismuth_blob_storage/")

    def create(self, key: str, value: bytes) -> None:
        path = self._dir / key
        if path.exists():
            raise ValueError("Key already exists.")
        path.parent.mkdir(parents=True, exist_ok=True)
        with open(self._dir / key, 'wb') as f:
            f.write(value)

    def retrieve(self, key: str) -> Optional[bytes]:
        path = self._dir / key
        if not path.exists():
            return None
        with open(path, 'rb') as f:
            return f.read()

    def update(self, key: str, value: bytes) -> None:
        path = self._dir / key
        if not path.exists():
            raise ValueError("Key does not exist.")
        with open(self._dir / key, 'wb') as f:
            f.write(value)

    def delete(self, key: str) -> None:
        path = self._dir / key
        if not path.exists():
            raise ValueError("Key does not exist.")
        path.unlink()

    def list_keys(self) -> set[str]:
        out = set()
        for fn in self._dir.glob('**/*'):
            out.add(str(fn.relative_to(self._dir)))
        return out

    def clear(self) -> None:
        if self._dir.exists():
            shutil.rmtree(self._dir)



from .api import *
from .api_with_storage import *
from .base_code_block import *
from .blob_storage import *
from .configuration import *
from .data_storage import *
from .function import *
from .sql import *



from typing import Callable, Dict, List
from flask import request, Request
from flask_restx import Resource
from .api import API
from .data_storage import DataStorage


class APIWithStorage(API):
    """
    A class that extends APICodeBlock to include functionalities for data storage.
    It manages API routes and associates them with specific data storage operations.
    """
    # A dictionary that holds instances of DataStorageCodeBlock, used for managing data storage.
    data_stores: Dict[str, DataStorage]

    def __init__(self, *args, **kwargs):
        """
        Initializes the APICodeBlockWithStorage instance.
        Inherits initialization from APICodeBlock and initializes a dictionary to hold DataStorageBlock instances.
        """
        super().__init__(*args, **kwargs)
        self.data_stores = {}

    def add_route_with_storage(
        self,
        route: str,
        methods: List[str],
        data_storage_block: DataStorage,
        require_auth: List[str] = ["POST", "DELETE", "PUT"],
        auth_cb: Callable[[Request], bool] = lambda r: True,
    ):
        """
        Adds a route to the API with associated data storage functionality.
        """
        self.data_stores[route] = data_storage_block
        methods = methods if isinstance(methods, list) else [methods]
        methods = [method.upper() for method in methods]

        class DynamicResource(Resource):
            def get(self):
                if "GET" in methods:
                    if "GET" in require_auth and not auth_cb(request):
                        return self.api.abort(401)

                    key = request.args.get("key")
                    return (
                        data_storage_block.retrieve(key)
                        if key
                        else data_storage_block.list_all()
                    )
                else:
                    self.api.abort(405)

            def post(self):
                if "POST" in methods:
                    if "POST" in require_auth and not auth_cb(request):
                        return self.api.abort(401)

                    data = request.json
                    key, value = data.get("key"), data.get("value")
                    if key and value:
                        data_storage_block.create(key, value)
                        return {"message": "Item created"}, 201
                    else:
                        return {"message": "Key and value required"}, 400
                else:
                    self.api.abort(405)

            def put(self):
                if "PUT" in methods:
                    if "PUT" in require_auth and not auth_cb(request):
                        return self.api.abort(401)

                    data = request.json
                    key, value = data.get("key"), data.get("value")
                    if key and value:
                        data_storage_block.update(key, value)
                        return {"message": "Item updated"}
                    else:
                        return {"message": "Key and value required"}, 400
                else:
                    self.api.abort(405)

            def delete(self):
                self._auth_handler()
                if "DELETE" in methods:
                    if "DELETE" in require_auth and not auth_cb(request):
                        return self.api.abort(401)

                    key = request.args.get("key")
                    if key:
                        data_storage_block.delete(key)
                        return {"message": "Item deleted"}
                    else:
                        return {"message": "Key required"}, 400

        self.api.add_resource(DynamicResource, route)
